import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:keframe/frame_separate_widget.dart';

import '../item/complex_item.dart';

class ComplexListExample extends StatefulWidget {
  @override
  _ComplexListExampleState createState() => _ComplexListExampleState();
}

class _ComplexListExampleState extends State<ComplexListExample> {
  int childCount = 100;

  ScrollController scrollController;
  double scrollPos = 1500;

  @override
  void initState() {
    super.initState();
    scrollController = ScrollController();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      resizeToAvoidBottomInset: true,
      appBar: AppBar(
        title: Text('listView'),
      ),
      body: Column(
        children: <Widget>[
          Expanded(
            child: FutureBuilder<int>(
              future: Future.delayed(Duration(seconds: 1), () => 100),
              // 模拟加载数据耗时1秒
              initialData: 0,
              builder: (BuildContext context, AsyncSnapshot<int> snapshot) {
                return ListView.builder(
                  cacheExtent: 500,
                  controller: scrollController,
                  itemCount: snapshot.data,
                  itemBuilder: (c, i) => FrameSeparateWidget(
                    index: i,
                    placeHolder: Container(
                      color: i % 2 == 0 ? Colors.yellow : Colors.green,
                      // 修改占位的背景色
                      height: 60,
                    ),
                    child: CellWidget(
                      color: i % 2 == 0 ? Colors.red : Colors.blue,
                      index: i,
                    ),
                  ),
                );
              },
            ),
          ),
          operateBar()
        ],
      ),
    );
  }

  Widget operateBar() {
    return Container(
      height: 200,
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        mainAxisAlignment: MainAxisAlignment.start,
        children: <Widget>[
          Expanded(
            child: Wrap(
              children: <Widget>[
                RaisedButton(
                    onPressed: () {
                      childCount += 20;
                      setState(() {});
                    },
                    child: Text(
                      "setState增加20",
                      style: TextStyle(fontSize: 14),
                    )),
                RaisedButton(
                    onPressed: () {
                      setState(() {});
                      scrollController.animateTo(scrollPos,
                          duration: Duration(
                              milliseconds: scrollPos == 0 ? 1500 : 600),
                          curve: Curves.easeInOut);
                      scrollPos = scrollPos >= 6000 ? 0 : scrollPos + 1500;
                    },
                    child: Text(
                      "滚动到$scrollPos位置",
                      style: TextStyle(fontSize: 14),
                    )),
              ],
            ),
          )
        ],
      ),
    );
  }
}
